from django.http import HttpResponsePermanentRedirect, Http404
from django.views.generic import ListView, DetailView
from django.db.models import get_model

from oscar.apps.catalogue.signals import product_viewed, product_search

product_model = get_model('catalogue','product')
category_model = get_model('catalogue', 'category')


class ProductDetailView(DetailView):
    context_object_name = 'product'
    model = product_model
    view_signal = product_viewed
    template_folder = "catalogue"
    _product = None
    
    def get_object(self):
        if not self._product:
            self._product = super(ProductDetailView, self).get_object()
        return self._product
    
    def get(self, request, **kwargs):
        """
        Ensure that the correct URL is used
        """
        product = self.get_object()
        correct_path = product.get_absolute_url() 
        if correct_path != request.path:
            return HttpResponsePermanentRedirect(correct_path)
        response = super(ProductDetailView, self).get(request, **kwargs)
        
        # Send signal to record the view of this product
        self.view_signal.send(sender=self, product=product, user=request.user, request=request, response=response)
        return response;

    def get_template_names(self):
        """
        Return a list of possible templates.
        
        We try 2 options before defaulting to catalogue/detail.html:
        1). detail-for-upc-<upc>.html
        2). detail-for-class-<classname>.html
        
        This allows alternative templates to be provided for a per-product
        and a per-item-class basis.
        """    
        product = self.get_object()
        names = ['%s/detail-for-upc-%s.html' % (self.template_folder, product.upc), 
                 '%s/detail-for-class-%s.html' % (self.template_folder, product.product_class.name.lower()),
                 '%s/detail.html' % (self.template_folder)]
        return names


class ProductCategoryView(ListView):
    u"""A list of products"""
    context_object_name = "products"
    template_name = 'catalogue/browse.html'
    paginate_by = 20
    
    def get_categories(self):
        slug = self.kwargs['category_slug']
        try:
            category = category_model.objects.get(slug=slug)
        except category_model.DoesNotExist:
            raise Http404()
        categories = list(category.get_descendants())
        categories.append(category)
        return categories
    
    def get_context_data(self, **kwargs):
        context = super(ProductCategoryView, self).get_context_data(**kwargs)

        categories = self.get_categories()

        context['categories'] = categories
        context['category'] = categories[-1]
        context['summary'] = categories[-1].name
        return context    

    def get_queryset(self):
        return product_model.browsable.filter(categories__in=self.get_categories()).distinct()


class ProductListView(ListView):
    u"""A list of products"""
    context_object_name = "products"
    template_name = 'catalogue/browse.html'
    paginate_by = 20
    search_signal = product_search

    def get_search_query(self):
        u"""Return a search query from GET"""
        q = None
        if 'q' in self.request.GET and self.request.GET['q']:
            q = self.request.GET['q'].strip()
        return q

    def get_queryset(self):
        u"""Return a set of products"""
        q = self.get_search_query()
        if q:
            # Send signal to record the view of this product
            self.search_signal.send(sender=self, query=q, user=self.request.user)
            return product_model.browsable.filter(title__icontains=q)
        else:
            return product_model.browsable.all()
        
    def get_context_data(self, **kwargs):
        context = super(ProductListView, self).get_context_data(**kwargs)
        q = self.get_search_query()
        if not q:
            context['summary'] = 'All products'
        else:
            context['summary'] = "Products matching '%s'" % q
            context['search_term'] = q
        return context



class ReportGenerator(object):
    u"""
    Top-level class that needs to be subclassed to provide a 
    report generator.
    """
    
    filename_template = 'report-%s-to-%s.csv'
    mimetype = 'text/csv'
    code = ''
    description = '<insert report description>'
    
    def __init__(self, **kwargs):
        if 'start_date' in kwargs and 'end_date' in kwargs:
            self.start_date = kwargs['start_date']
            self.end_date = kwargs['end_date']
    
    def generate(self, response):
        pass
 
    def filename(self):
        u"""
        Returns the filename for this report
        """
        return self.filename_template % (self.start_date, self.end_date)
    
    def is_available_to(self, user):
        u"""
        Checks whether this report is available to this user
        """
        return user.is_staff


from django.contrib import messages
from django.core.urlresolvers import reverse
from django.db.models import get_model
from django.http import HttpResponseRedirect, Http404
from django.views.generic import ListView, FormView, View
from django.forms.models import modelformset_factory
from django.utils.translation import ugettext_lazy as _
from django.core.exceptions import ObjectDoesNotExist

from extra_views import ModelFormSetView
from oscar.apps.basket.forms import BasketLineForm, AddToBasketForm, \
    BasketVoucherForm, SavedLineForm
from oscar.apps.basket.signals import basket_addition
from oscar.core.loading import import_module
import_module('offer.utils', ['Applicator'], locals())


class BasketView(ModelFormSetView):
    model = get_model('basket', 'Line')
    basket_model = get_model('basket', 'Basket')
    form_class = BasketLineForm
    extra = 0
    can_delete = True
    template_name='basket/basket.html'
    
    def get_queryset(self):
        return self.request.basket.lines.all()
    
    def get_context_data(self, **kwargs):
        context = super(BasketView, self).get_context_data(**kwargs)
        context['voucher_form'] = BasketVoucherForm()
        
        if self.request.user.is_authenticated():
            try:
                saved_basket = self.basket_model.saved.get(owner=self.request.user)
                saved_queryset = saved_basket.lines.all().select_related('product', 'product__stockrecord')
                SavedFormset = modelformset_factory(self.model, form=SavedLineForm, extra=0, can_delete=True)
                formset = SavedFormset(queryset=saved_queryset)
                context['saved_formset'] = formset
            except self.basket_model.DoesNotExist:
                pass
        return context
    
    def get_success_url(self):
        return self.request.META.get('HTTP_REFERER', reverse('basket:summary'))

    def formset_valid(self, formset):
        needs_auth = False
        for form in formset:
            if form.cleaned_data['save_for_later']:
                line = form.instance
                if self.request.user.is_authenticated():
                    self.move_line_to_saved_basket(line)
                    messages.info(self.request, _(u"'%(title)s' has been saved for later" % {'title': line.product}))   
                else:
                    needs_auth = True
        if needs_auth:
            messages.error(self.request, "You can't save an item for later if you're not logged in!")     
        return super(BasketView, self).formset_valid(formset)

    def move_line_to_saved_basket(self, line):
        saved_basket, _ = get_model('basket', 'basket').saved.get_or_create(owner=self.request.user)
        saved_basket.merge_line(line)

    def formset_invalid(self, formset):
        messages.info(self.request, _("There was a problem updating your basket, please check that all quantities are numbers"))
        return super(BasketView, self).formset_invalid(formset)


class BasketAddView(FormView):
    """
    Handles the add-to-basket operation, shouldn't be accessed via 
    GET because there's nothing sensible to render.
    """
    form_class = AddToBasketForm
    product_model = get_model('catalogue', 'product')
    add_signal = basket_addition
    
    def get(self, request, *args, **kwargs):
        return HttpResponseRedirect(reverse('basket:summary'))
    
    def get_form_kwargs(self): 
        kwargs = super(BasketAddView, self).get_form_kwargs()
        
        try:
            product = self.product_model.objects.get(pk=self.request.POST['product_id'])
        except self.product_model.DoesNotExist:
            raise Http404()
        kwargs['instance'] = product
        return kwargs
    
    def get_success_url(self):
        return self.request.META.get('HTTP_REFERER', reverse('basket:summary'))

    def form_valid(self, form):
        options = []
        for option in form.instance.options:
            if option.code in form.cleaned_data:
                options.append({'option': option, 'value': form.cleaned_data[option.code]})
        self.request.basket.add_product(form.instance, form.cleaned_data['quantity'], options)
        messages.info(self.request, _(u"'%(title)s' (quantity %(quantity)d) has been added to your basket" %
                {'title': form.instance.get_title(), 
                 'quantity': form.cleaned_data['quantity']}))
        
        # Send signal for basket addition
        self.add_signal.send(sender=self, product=form.instance, user=self.request.user)
        
        return super(BasketAddView, self).form_valid(form)
    
    def form_invalid(self, form):
        msgs = []
        for error in form.errors.values():
            msgs.append(error.as_text()) 
        messages.error(self.request, ",".join(msgs))
        return HttpResponseRedirect(self.request.META.get('HTTP_REFERER',reverse('basket:summary')))


class VoucherView(ListView):
    model = get_model('voucher', 'voucher')
    can_delete = True
    extra = 0
    
    def get_queryset(self):
        self.request.basket.vouchers.all()


class VoucherAddView(FormView):
    form_class = BasketVoucherForm
    voucher_model = get_model('voucher', 'voucher')
    
    def get(self, request, *args, **kwargs):
        return HttpResponseRedirect(reverse('basket:summary'))
    
    def apply_voucher_to_basket(self, voucher):
        if not voucher.is_active():
            messages.error(self.request, _("The '%(code)s' voucher has expired" % {'code': voucher.code}))
            return
        
        is_available, message = voucher.is_available_to_user(self.request.user)
        if not is_available:
            messages.error(self.request, message)
            return
        
        self.request.basket.vouchers.add(voucher)
        
        # Recalculate discounts to see if the voucher gives any
        discounts_before = self.request.basket.get_discounts()
        self.request.basket.remove_discounts()
        Applicator().apply(self.request, self.request.basket)
        discounts_after = self.request.basket.get_discounts()
        
        # Look for discounts from this new voucher
        found_discount = False
        for discount in discounts_after:
            if discount['voucher'] and discount['voucher'] == voucher:
                found_discount = True
                break
        if not found_discount:
            messages.warning(self.request, _("Your basket does not qualify for a voucher discount"))
            self.request.basket.vouchers.remove(voucher)
        else:
            messages.info(self.request, _("Voucher '%(code)s' added to basket" % {'code': voucher.code}))
    
    def form_valid(self, form):
        code = form.cleaned_data['code']
        if self.request.basket.contains_voucher(code):
            messages.error(self.request, _("You have already added the '%(code)s' voucher to your basket" % {'code': code}))
        else:
            try:
                voucher = self.voucher_model._default_manager.get(code=code)
            except self.voucher_model.DoesNotExist:
                messages.error(self.request, _("No voucher found with code '%(code)s'" % {'code': code}))
            else:        
                self.apply_voucher_to_basket(voucher)
        return HttpResponseRedirect(self.request.META.get('HTTP_REFERER', reverse('basket:summary')))

    def form_invalid(self, form):
        return HttpResponseRedirect(reverse('basket:summary'))
    

class VoucherRemoveView(View):
    voucher_model = get_model('voucher', 'voucher')
    
    def get(self, request, *args, **kwargs):
        return HttpResponseRedirect(reverse('basket:summary'))
    
    def post(self, request, *args, **kwargs):
        voucher_id = int(kwargs.pop('pk'))
        try:
            voucher = request.basket.vouchers.get(id=voucher_id)
        except ObjectDoesNotExist:
            messages.error(request, "No voucher found with id '%d'" % voucher_id)
        else:
            request.basket.vouchers.remove(voucher)
            request.basket.save()
            messages.info(request, "Voucher '%s' removed from basket" % voucher.code)
        return HttpResponseRedirect(reverse('basket:summary'))


class SavedView(ModelFormSetView):
    model = get_model('basket', 'line')
    basket_model = get_model('basket', 'basket')
    form_class = SavedLineForm
    extra = 0
    can_delete = True

    def get(self, request, *args, **kwargs):
        return HttpResponseRedirect(reverse('basket:summary'))

    def get_queryset(self):
        try:
            saved_basket = self.basket_model.saved.get(owner=self.request.user)
            return saved_basket.lines.all().select_related('product', 'product__stockrecord')
        except self.basket_model.DoesNotExist:
            return []
        
    def get_success_url(self):
        return self.request.META.get('HTTP_REFERER', reverse('basket:summary'))

    def formset_valid(self, formset):
        for form in formset:
            if form.cleaned_data['move_to_basket']:
                msg = "'%s' has been moved back to your basket" % form.instance.product
                messages.info(self.request, msg)                
                real_basket = self.request.basket
                real_basket.merge_line(form.instance)
        return super(SavedView, self).formset_valid(formset)

    def formset_invalid(self, formset):
        return HttpResponseRedirect(self.request.META.get('HTTP_REFERER', reverse('basket:summary')))



#    def do_bulk_load(self, basket):
#        num_additions = 0
#        num_not_found = 0
#        for sku in re.findall(r"[\d -]{5,}", self.request.POST['source_text']):
#            try:
#                item = Product.objects.get(upc=sku)
#                basket.add_product(item)
#                num_additions += 1
#            except Product.DoesNotExist:
#                num_not_found += 1
#        messages.info(self.request, "Added %d items to your basket (%d missing)" % (num_additions, num_not_found))
#



from django.test import TestCase, Client
from django.core.exceptions import ValidationError
from django.core.urlresolvers import reverse

from oscar.apps.catalogue.models import Product, ProductClass, Category
from oscar.apps.catalogue.utils import breadcrumbs_to_category


class CategoryTests(TestCase):
    
    def test_create_category_root(self):
        trail = 'Books'
        category = breadcrumbs_to_category(trail)
        self.assertIsNotNone(category)
        self.assertEquals(category.name, 'Books')
        self.assertEquals(category.slug, 'books')      
    
    def test_subcategory(self):
        trail = 'Books > Science-Fiction'
        category = breadcrumbs_to_category(trail)
        
        self.assertIsNotNone(category)
        self.assertEquals(category.name, 'Science-Fiction')
        self.assertEquals(category.get_depth(), 2)
        self.assertEquals(category.get_parent().name, 'Books')
        self.assertEquals(2, Category.objects.count())
        self.assertEquals(category.slug, 'books/science-fiction')
        
    def test_subsubcategory(self):
        trail = 'Books > Science-Fiction > Star Trek'
        breadcrumbs_to_category(trail)
        trail = 'Books > Factual > Popular Science'
        category = breadcrumbs_to_category(trail)        
        
        self.assertIsNotNone(category)
        self.assertEquals(category.name, 'Popular Science')
        self.assertEquals(category.get_depth(), 3)
        self.assertEquals(category.get_parent().name, 'Factual')        
        self.assertEquals(5, Category.objects.count())
        self.assertEquals(category.slug, 'books/factual/popular-science', )        


class ItemTests(TestCase):

    def setUp(self):
        self.product_class,_ = ProductClass.objects.get_or_create(name='Clothing')
   

class TopLevelItemTests(ItemTests):
    
    def test_top_level_products_must_have_titles(self):
        self.assertRaises(ValidationError, Product.objects.create, product_class=self.product_class)
        
        
class VariantItemTests(ItemTests):
    
    def setUp(self):
        super(VariantItemTests, self).setUp()
        self.parent = Product.objects.create(title="Parent product", product_class=self.product_class)
    
    def test_variant_products_dont_need_titles(self):
        p = Product.objects.create(parent=self.parent, product_class=self.product_class)
        
    def test_variant_products_dont_need_a_product_class(self):
        p = Product.objects.create(parent=self.parent)
        
    def test_variant_products_inherit_parent_titles(self):
        p = Product.objects.create(parent=self.parent, product_class=self.product_class)
        self.assertEquals("Parent product", p.get_title())
        
    def test_variant_products_inherit_product_class(self):
        p = Product.objects.create(parent=self.parent)
        self.assertEquals("Clothing", p.get_product_class().name)


class SingleProductViewTest(TestCase):
    fixtures = ['sample-products']
    
    def setUp(self):
        self.client = Client()
        
    def test_canonical_urls_are_enforced(self):
        p = Product.objects.get(id=1)
        args = {'product_slug': 'wrong-slug',
                'pk': p.id}
        wrong_url = reverse('catalogue:detail', kwargs=args)
        response = self.client.get(wrong_url)
        self.assertEquals(301, response.status_code)



from decimal import Decimal as D

from django.conf import settings
from django.db import models
from django.utils.translation import ugettext_lazy as _
from django.utils.importlib import import_module as django_import_module

from oscar.core.loading import import_module
import_module('partner.wrappers', ['DefaultWrapper'], locals())


# Cache the partners for quicklookups
default_wrapper = DefaultWrapper()
partner_wrappers = {}
for partner, class_str in settings.OSCAR_PARTNER_WRAPPERS.items():
    bits = class_str.split('.')
    class_name = bits.pop()
    module_str = '.'.join(bits)
    module = django_import_module(module_str)
    partner_wrappers[partner] = getattr(module, class_name)()

def get_partner_wrapper(partner_name):
    """
    Returns the appropriate partner wrapper given the partner name
    """
    return partner_wrappers.get(partner_name, default_wrapper)


class AbstractPartner(models.Model):
    u"""Fulfillment partner"""
    name = models.CharField(max_length=128, unique=True)
    
    # A partner can have users assigned to it.  These can be used
    # to provide authentication for webservices etc.
    users = models.ManyToManyField('auth.User', related_name="partners", blank=True, null=True)
    
    class Meta:
        verbose_name_plural = 'Fulfillment partners'
        abstract = True
        permissions = (
            ("can_edit_stock_records", "Can edit stock records"),
            ("can_view_stock_records", "Can view stock records"),
            ("can_edit_product_range", "Can edit product range"),
            ("can_view_product_range", "Can view product range"),
            ("can_edit_order_lines", "Can edit order lines"),
            ("can_view_order_lines", "Can view order lines"),
        )
        
    def __unicode__(self):
        return self.name


class AbstractStockRecord(models.Model):
    """
    A basic stock record.
    
    This links a product to a partner, together with price and availability
    information.  Most projects will need to subclass this object to add custom
    fields such as lead_time, report_code, min_quantity.
    """
    product = models.OneToOneField('catalogue.Product', related_name="stockrecord")
    partner = models.ForeignKey('partner.Partner')
    
    # The fulfilment partner will often have their own SKU for a product, which
    # we store here.
    partner_sku = models.CharField(_("Partner SKU"), max_length=128, blank=True)
    
    # Price info:
    # We deliberately don't store tax information to allow each project
    # to subclass this model and put its own fields for convey tax.
    price_currency = models.CharField(max_length=12, default=settings.OSCAR_DEFAULT_CURRENCY)
    
    # This is the base price for calculations - tax should be applied 
    # by the appropriate method.  We don't store it here as its calculation is 
    # highly domain-specific.  It is NULLable because some items don't have a fixed
    # price.
    price_excl_tax = models.DecimalField(decimal_places=2, max_digits=12, blank=True, null=True)
    
    # Retail price for this item
    price_retail = models.DecimalField(decimal_places=2, max_digits=12, blank=True, null=True)
    
    # Cost price is optional as not all partner supply it
    cost_price = models.DecimalField(decimal_places=2, max_digits=12, blank=True, null=True)
    
    # Stock level information
    num_in_stock = models.IntegerField(default=0, blank=True, null=True)
    
    # The amount of stock allocated to orders but not fed back to the master
    # stock system.  A typical stock update process will set the num_in_stock
    # variable to a new value and reset num_allocated to zero
    num_allocated = models.IntegerField(default=0, blank=True, null=True)
    
    # Date information
    date_created = models.DateTimeField(auto_now_add=True)
    date_updated = models.DateTimeField(auto_now=True, db_index=True)
    
    class Meta:
        abstract = True
    
    def allocate(self, quantity):
        """
        Record a stock allocation.
        
        We don't alter the num_in_stock variable as it is assumed that this
        will be set by a batch "stock update" process.
        """
        self.num_allocated = int(self.num_allocated)
        self.num_allocated += quantity
        self.save()
        
    def set_discount_price(self, price):
        """
        A setter method for setting a new price.  
        
        This is called from within the "discount" app, which is responsible
        for applying fixed-discount offers to products.  We use a setter method
        so that this behaviour can be customised in projects.
        """
        self.price_excl_tax = price
        self.save()
        
    # Price retrieval methods - these default to no tax being applicable
    # These are intended to be overridden.   
    
    @property
    def is_available_to_buy(self):
        """
        Return whether this stockrecord allows the product to be purchased
        """
        return get_partner_wrapper(self.partner.name).is_available_to_buy(self)
    
    @property
    def net_stock_level(self):
        """
        Return the effective number in stock
        """ 
        if self.num_in_stock is None:
            return 0
        if self.num_allocated is None:
            return self.num_in_stock
        return self.num_in_stock - self.num_allocated
    
    @property
    def availability(self):
        """
        Return an item's availability as a string
        """
        return get_partner_wrapper(self.partner.name).availability(self)
    
    @property
    def dispatch_date(self):
        """
        Return the estimated dispatch date for a line
        """
        return get_partner_wrapper(self.partner.name).dispatch_date(self)
    
    @property
    def lead_time(self):
        return get_partner_wrapper(self.partner.name).lead_time(self)
    
    @property 
    def price_incl_tax(self):
        """
        Return a product's price including tax.
        
        This defaults to the price_excl_tax as tax calculations are 
        domain specific.  This class needs to be subclassed and tax logic
        added to this method.
        """
        if self.price_excl_tax is None:
            return D('0.00')
        return self.price_excl_tax + self.price_tax
    
    @property 
    def price_tax(self):
        """
        Return a product's tax value
        """
        return get_partner_wrapper(self.partner.name).calculate_tax(self)
    
    def __unicode__(self):
        if self.partner_sku:
            return "%s (%s): %s" % (self.partner.name, self.partner_sku, self.product.title)
        else:
            return "%s: %s" % (self.partner.name, self.product.title)



from decimal import Decimal

from django.db import models
from django.template.defaultfilters import slugify
from django.utils.translation import ugettext as _
from django.conf import settings


class Transaction(models.Model):
    """
    A transaction for payment sources which need a secondary 'transaction' to actually take the money
    
    This applies mainly to credit card sources which can be a pre-auth for the money.  A 'complete'
    needs to be run later to debit the money from the account.
    """
    source = models.ForeignKey('payment.Source', related_name='transactions')
    
    # We define some sample types
    AUTHORISE, DEBIT, REFUND = 'Authorise', 'Debit', 'Refund'
    txn_type = models.CharField(max_length=128, blank=True)
    amount = models.DecimalField(decimal_places=2, max_digits=12)
    reference = models.CharField(max_length=128, null=True)
    status = models.CharField(max_length=128, null=True)
    date_created = models.DateTimeField(auto_now_add=True)
    
    def __unicode__(self):
        return "%s of %.2f" % (self.txn_type, self.amount)


class Source(models.Model):
    """
    A source of payment for an order.  
    
    This is normally a credit card which has been pre-authed for the order
    amount, but some applications will allow orders to be paid for using
    multiple sources such as cheque, credit accounts, gift cards.  Each payment
    source will have its own entry.
    """
    order = models.ForeignKey('order.Order', related_name='sources')
    source_type = models.ForeignKey('payment.SourceType')
    currency = models.CharField(max_length=12, default=settings.OSCAR_DEFAULT_CURRENCY)
    
    # Track the various amounts associated with this source
    amount_allocated = models.DecimalField(decimal_places=2, max_digits=12, default=Decimal('0.00'))
    amount_debited = models.DecimalField(decimal_places=2, max_digits=12, default=Decimal('0.00'))
    amount_refunded = models.DecimalField(decimal_places=2, max_digits=12, default=Decimal('0.00'))
    
    # Reference number for this payment source.  This is often used to look up a
    # transaction model for a particular payment partner. 
    reference = models.CharField(max_length=128, blank=True, null=True)
    
    # A customer-friendly label for the source, eg XXXX-XXXX-XXXX-1234
    label = models.CharField(max_length=128, blank=True, null=True)
    
    # A dictionary of submission data that is stored as part of the
    # checkout process.
    submission_data = None
    
    # We keep a list of deferred transactions that are only actually saved when 
    # the source is saved for the first time
    deferred_txns = None
    
    def __unicode__(self):
        description = "Allocation of %.2f from type %s" % (self.amount_allocated, self.source_type)
        if self.reference:
            description += " (reference: %s)" % self.reference
        return description
    
    def save(self, *args, **kwargs):
        super(Source, self).save(*args, **kwargs)
        if self.deferred_txns:
            for txn in self.deferred_txns:
                self._create_transaction(*txn)
    
    def balance(self):
        return self.amount_allocated - self.amount_debited + self.amount_refunded
    
    def create_deferred_transaction(self, txn_type, amount, reference=None, status=None):
        """
        Register the data for a transaction that can't be created yet due to FK
        constraints.  This happens at checkout where create an payment source and a
        transaction but can't save them until the order model exists.
        """
        if self.deferred_txns is None:
            self.deferred_txns = []
        self.deferred_txns.append((txn_type, amount, reference, status))
    
    def _create_transaction(self, txn_type, amount, reference=None, status=None):
        Transaction.objects.create(source=self,
                                   txn_type=txn_type,
                                   amount=amount,
                                   reference=reference,
                                   status=status)
    
    def allocate(self, amount, reference=None, status=None):
        """
        Convenience method for ring-fencing money against this source
        """
        self.amount_allocated += amount
        self.save()
        self._create_transaction(Transaction.AUTHORISE, amount, reference, status)
    
    def debit(self, amount=None, reference=None, status=None):
        """
        Convenience method for recording debits against this source
        """
        if amount is None:
            amount = self.balance()
        self.amount_debited += amount
        self.save()
        self._create_transaction(Transaction.DEBIT, amount, reference, status)
        
    def refund(self, amount, reference=None, status=None):
        """
        Convenience method for recording refunds against this source
        """
        self.amount_refunded += amount
        self.save()
        self._create_transaction(Transaction.REFUND, amount, reference, status)
    
    @property
    def amount_available_for_refund(self):
        """
        Return the amount available to be refunded
        """
        return self.amount_debited - self.amount_refunded
    
    
class SourceType(models.Model):
    """
    A type of payment source.
    
    This could be an external partner like PayPal or DataCash,
    or an internal source such as a managed account.i
    """
    name = models.CharField(max_length=128)
    code = models.SlugField(max_length=128, help_text=_("""This is used within
        forms to identify this source type"""))

    def __unicode__(self):
        return self.name

    def save(self, *args, **kwargs):
        if not self.code:
            self.code = slugify(self.name)
        super(SourceType, self).save(*args, **kwargs)
    

class Bankcard(models.Model):
    user = models.ForeignKey('auth.User', related_name='bankcards')
    card_type = models.CharField(max_length=128)
    name = models.CharField(max_length=255)
    number = models.CharField(max_length=32)
    expiry_date = models.DateField()
    
    # For payment partners who are storing the full card details for us
    partner_reference = models.CharField(max_length=255, null=True, blank=True)
    
    def save(self, *args, **kwargs):
        self.number = self._get_obfuscated_number()
        super(Bankcard, self).save(*args, **kwargs)    
        
    def _get_obfuscated_number(self):
        return u"XXXX-XXXX-XXXX-%s" % self.number[-4:]



from decimal import Decimal, ROUND_UP

from django.db import models
from django.utils.translation import ugettext as _

ABSOLUTE_DISCOUNT, PERCENTAGE_DISCOUNT, FINAL_PRICE = "Absolute", "Percentage", "Final price"

class AbstractDiscountOffer(models.Model):
    u"""
    A fixed-discount offer (eg get 10% off all fiction books)
    """
    name = models.CharField(max_length=128)
    description = models.TextField(blank=True, null=True)
    
    start_date = models.DateField()
    end_date = models.DateField()
    
    products = models.ManyToManyField('catalogue.Product')
    TYPE_CHOICES = (
        (ABSOLUTE_DISCOUNT, _("An absolute amount of discount off the site price")),
        (PERCENTAGE_DISCOUNT, _("A percentage discount off the site price")),
        (FINAL_PRICE, _("Sets the site price")),
    )
    discount_type = models.CharField(_("Discount type"), max_length=128, choices=TYPE_CHOICES, default=PERCENTAGE_DISCOUNT)
    discount_value = models.DecimalField(decimal_places=2, max_digits=12, 
        help_text="""For percentage values, enter integers less than 100 (eg '25' for a 25% discount)""")
    
    date_created = models.DateTimeField(auto_now_add=True)

    class Meta:
        ordering = ['-date_created']
        abstract = True
        
    def __unicode__(self):
        return self.name    
        
    def is_active(self, test_date=None):
        u"""
        Tests whether this offer is currently active or not.
        """
        if not test_date:
            test_date = datetime.date.today()
        return self.start_date <= test_date and test_date < self.end_date

    def num_products(self):
        return self.products.count()

    def apply(self):
        u"""
        Applies this offer to its linked products.
        
        The default behaviour of this method is to always
        apply this offer, even if it makes the product more expensive.
        This behaviour can of course customised by subclassing.
        """
        for product in self.products.all():
            self._apply_discount_to_product(product)
            
    def _apply_discount_to_product(self, product):
        u"""
        Applies this offer to an individual product
        """
        if product.has_stockrecord:
            discount_price = self._get_discount_price(product)
            product.stockrecord.set_discount_price(discount_price)
    
    def _get_discount_price(self, product):
        u"""
        Returns the discounted price
        """
        current_price = product.stockrecord.price_excl_tax
        if self.discount_type == ABSOLUTE_DISCOUNT:
            price = max(Decimal('0.00'), current_price - self.discount_value)
        elif self.discount_type == PERCENTAGE_DISCOUNT:
            price = current_price * (1 - self.discount_value/100)
        else:
            price = self.discount_value
        return price.quantize(Decimal('.01'))
                

        
        



from oscar.apps.address.tests import *
from oscar.apps.basket.tests import *
from oscar.apps.order.tests import *
from oscar.apps.catalogue.tests import *
from oscar.apps.partner.tests import *
from oscar.apps.checkout.tests import *
from oscar.apps.payment.tests import *
from oscar.apps.offer.tests import *
from oscar.apps.shipping.tests import *
from oscar.apps.customer.tests import *
from oscar.apps.discount.tests import *
from oscar.apps.promotions.tests import *
from oscar.apps.catalogue.reviews.tests import *
from oscar.apps.voucher.tests import *
from oscar.apps.partner.tests import *

from oscar.core.tests import *
from oscar.core.logging.tests import *



import os
import csv
import sys
from decimal import Decimal as D

from oscar.core.loading import import_module

import_module('partner.exceptions', ['ImportError'], locals())
import_module('partner.models', ['Partner', 'StockRecord'], locals())
import_module('catalogue.models', ['ProductClass', 'Product'], locals())


class StockImporter(object):
    
    def __init__(self, logger, partner, delimiter):
        self.logger = logger
        self._delimiter = delimiter
        
        try:
            self._partner = Partner.objects.get(name=partner)
        except Partner.DoesNotExist:
            name_list = ", ".join([d['name'] for d in Partner.objects.values('name')])
            raise ImportError("Partner named '%s' does not exist (existing partners: %s)" % (partner, name_list))
        
    def handle(self, file_path=None):
        u"""Handles the actual import process"""
        if not file_path:
            raise ImportError("No file path supplied")
        Validator().validate(file_path)
        self._import(file_path)

    def _import(self, file_path):
        u"""Imports given file"""
        stats = {'updated_items': 0,
                 'unchanged_items': 0,
                 'unmatched_items': 0}
        row_number = 0
        for row in csv.reader(open(file_path, 'rb'), delimiter=self._delimiter, quotechar='"', escapechar='\\'):
            row_number += 1
            self._import_row(row_number, row, stats)
        msg = "\tUpdated items: %d\n\tUnchanged items: %d\n\tUnmatched items: %d" % (
            stats['updated_items'], 
            stats['unchanged_items'],
            stats['unmatched_items'])
        self.logger.info(msg)
    
    def _import_row(self, row_number, row, stats):
        if len(row) != 3:
            self.logger.error("Row number %d has an invalid number of fields, skipping..." % row_number)
        else:
            self._update_stockrecord(*row[:3], row_number=row_number, stats=stats)
    
    def _update_stockrecord(self, partner_sku, price_excl_tax, num_in_stock, row_number, stats):
        try:         
            stock = StockRecord.objects.get(partner=self._partner, partner_sku=partner_sku)
        except StockRecord.DoesNotExist:
            stats['unmatched_items'] += 1
            self.logger.error("\t - Row %d: StockRecord for partner '%s' and sku '%s' does not exist, skipping..." % (row_number, self._partner, partner_sku))
            return
        
        price_changed = False
        if stock.price_excl_tax != D(price_excl_tax):
            stock.price_excl_tax = D(price_excl_tax)
            price_changed = True
            
        stock_changed = False    
        if stock.num_in_stock != int(num_in_stock):
            stock.num_in_stock = num_in_stock
            stock_changed = True
        
        if price_changed or stock_changed:
            stock.save()
            
            msg = " SKU %s:" % (partner_sku)
            if price_changed:
                msg += '\n - Price set to %s' % (price_excl_tax)
            if stock_changed:
                msg += '\n - Stock set to %s' % num_in_stock
            self.logger.info(msg)            
            stats['updated_items'] += 1
        else:
            stats['unchanged_items'] += 1


class CatalogueImporter(object):
    u"""A catalogue importer object"""
    
    _flush = False
    
    def __init__(self, logger, delimiter=",", flush=False):
        self.logger = logger
        self._delimiter = delimiter
        self._flush = flush
        if flush:
            self.logger.info(" - Flushing product data before import")
    
    def handle(self, file_path=None):
        u"""Handles the actual import process"""
        if not file_path:
            raise ImportError("No file path supplied")
        Validator().validate(file_path)
        if self._flush is True:
            self._flush_product_data()
        self._import(file_path)
        
    def _flush_product_data(self):
        u"""Flush out product and stock models"""
        ProductClass.objects.all().delete()
        Product.objects.all().delete()
        Partner.objects.all().delete()
        StockRecord.objects.all().delete()

    def _import(self, file_path):
        u"""Imports given file"""
        stats = {'new_items': 0,
                 'updated_items': 0
                 }
        row_number = 0
        for row in csv.reader(open(file_path,'rb'), delimiter=self._delimiter, quotechar='"', escapechar='\\'):
            row_number += 1
            self._import_row(row_number, row, stats)
        msg = "\tNew items: %d\n\tUpdated items: %d" % (stats['new_items'], stats['updated_items'])
        self.logger.info(msg)
    
    def _import_row(self, row_number, row, stats):
        if len(row) != 4 and len(row) != 8:
            self.logger.error("Row number %d has an invalid number of fields (%d), skipping..." % (row_number, len(row)))
            return
        item = self._create_item(*row[:4], stats=stats)
        if len(row) == 8:
            # With stock data
            self._create_stockrecord(item, *row[4:8], stats=stats)
            
    def _create_item(self, upc, title, description, product_class, stats):
        # Ignore any entries that are NULL
        if description == 'NULL':
            description = ''
        
        # Create item class and item
        product_class,_ = ProductClass.objects.get_or_create(name=product_class)
        try:
            item = Product.objects.get(upc=upc)
            stats['updated_items'] += 1
        except Product.DoesNotExist:
            item = Product()
            stats['new_items'] += 1
        item.upc = upc
        item.title = title
        item.description = description
        item.product_class = product_class
        item.save()
        return item
        
    def _create_stockrecord(self, item, partner_name, partner_sku, price_excl_tax, num_in_stock, stats):            
        # Create partner and stock record
        partner, _ = Partner.objects.get_or_create(name=partner_name)
        try:
            stock = StockRecord.objects.get(partner_sku=partner_sku)
        except StockRecord.DoesNotExist:
            stock = StockRecord()
        
        stock.product = item
        stock.partner = partner
        stock.partner_sku = partner_sku
        stock.price_excl_tax = D(price_excl_tax)
        stock.num_in_stock = num_in_stock
        stock.save()    
    
        
class Validator(object):
    
    def validate(self, file_path):
        self._exists(file_path)
        self._is_file(file_path)
        self._is_readable(file_path)
    
    def _exists(self, file_path):
        u"""Check whether a file exists"""
        if not os.path.exists(file_path):
            raise ImportError("%s does not exist" % (file_path))
        
    def _is_file(self, file_path):
        u"""Check whether file is actually a file type"""
        if not os.path.isfile(file_path):
            raise ImportError("%s is not a file" % (file_path))
        
    def _is_readable(self, file_path):
        u"""Check file is readable"""
        try:
            f = open(file_path, 'r')
            f.close()
        except:
            raise ImportError("%s is not readable" % (file_path))


from decimal import Decimal as D
from xml.dom.minidom import parseString
import datetime

from django.test import TestCase
from django.conf import settings

from oscar.apps.payment.datacash.models import OrderTransaction
from oscar.apps.payment.datacash.utils import Gateway, Facade
from oscar.apps.payment.utils import Bankcard


class OrderTransactionTests(TestCase):
    
    def test_cc_numbers_are_not_saved(self):
        
        request_xml = """<?xml version="1.0" encoding="UTF-8" ?>
<Request>
    <Authentication>
        <client>99000001</client>
        <password>boomboom</password>
    </Authentication>
    <Transaction>
    <CardTxn>
        <Card>
            <pan>1000011100000004</pan>
            <expirydate>04/06</expirydate>
            <startdate>01/04</startdate>
        </Card>
        <method>auth</method>
    </CardTxn>
    <TxnDetails>
        <merchantreference>1000001</merchantreference>
        <amount currency="GBP">95.99</amount>
    </TxnDetails>
    </Transaction>
</Request>"""

        response_xml = """<?xml version="1.0" encoding="UTF-8" ?>
<Response>
    <CardTxn>
        <authcode>060642</authcode>
        <card_scheme>Switch</card_scheme>
        <country>United Kingdom</country>
        <issuer>HSBC</issuer>
    </CardTxn>
    <datacash_reference>3000000088888888</datacash_reference>
    <merchantreference>1000001</merchantreference>
    <mode>LIVE</mode>
    <reason>ACCEPTED</reason>
    <status>1</status>
    <time>1071567305</time>
</Response>"""
        
        txn = OrderTransaction.objects.create(order_number='1000',
                                              method='auth',
                                              datacash_ref='3000000088888888',
                                              merchant_ref='1000001',
                                              amount=D('95.99'),
                                              status=1,
                                              reason='ACCEPTED',
                                              request_xml=request_xml,
                                              response_xml=response_xml)
        doc = parseString(txn.request_xml)
        element = doc.getElementsByTagName('pan')[0]
        self.assertEqual('XXXXXXXXXXXX0004', element.firstChild.data)
        
        
class IntegrationTests(TestCase):
    
    def _test_for_smoke(self):
        gateway = Gateway(settings.DATACASH_CLIENT, 
                          settings.DATACASH_PASSWORD,
                          host=settings.DATACASH_HOST)
        response = gateway.auth(card_number='1000011000000005',
                                expiry_date='01/13',
                                amount=D('50.00'),
                                currency='GBP',
                                merchant_reference='123456_%s' % datetime.datetime.now().microsecond)
        print response
        
    def _test_adapter(self):
        bankcard = Bankcard(card_number='1000011000000005', expiry_date='01/13')
    
        dc_facade = Facade()
        reference = dc_facade.debit('102910', D('23.00'), bankcard)
        print reference
        
        OrderTransaction.objects.get(order_number='102910')
        
    


from django.conf.urls.defaults import patterns, url, include

from oscar.core.application import Application
from oscar.apps.catalogue.app import application as catalogue_app
from oscar.apps.customer.app import application as customer_app
from oscar.apps.basket.app import application as basket_app
from oscar.apps.checkout.app import application as checkout_app
from oscar.apps.promotions.app import application as promotions_app
from oscar.apps.search.app import application as search_app


class Shop(Application):
    name = None
    
    catalogue_app = catalogue_app
    customer_app = customer_app
    basket_app = basket_app
    checkout_app = checkout_app
    promotions_app = promotions_app
    search_app = search_app
    
    def get_urls(self):
        urlpatterns = patterns('',
            (r'products/', include(self.catalogue_app.urls)),
            (r'basket/', include(self.basket_app.urls)),
            (r'checkout/', include(self.checkout_app.urls)),
            (r'accounts/', include(self.customer_app.urls)),
            (r'search/', include(self.search_app.urls)),
            (r'', include(self.promotions_app.urls)),             
        )
        return urlpatterns
    
shop = Shop()


from django.conf.urls.defaults import patterns


class Application(object):
    name = None
    
    def __init__(self, app_name=None, **kwargs):
        self.app_name = app_name
        # Set all kwargs as object attributes
        for key, value in kwargs.iteritems():
            setattr(self, key, value)
    
    def get_urls(self):
        """
        Return the url patterns for this app, MUST be implemented in the subclass
        """
        return patterns('')
    
    @property
    def urls(self):
        # We set the application and instance namespace here
        return self.get_urls(), self.app_name, self.name